/** 
 *   Thermometer logger novus
 *
 *   (c) Copyright 2005 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/



#define _GNU_SOURCE

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <sys/file.h>
#include <string.h>

#include "my_stdio.h"
#include "modbus.h"


static ssize_t soft_read_time_out(struct timeval tv, int fd, void *buf, 
				  size_t count)
{
	fd_set rfds;
	int fd_max = fd + 1;
	int retval, l = 0;
	struct timeval tvl;

	while ( l < count) {
	  memcpy(&tvl, &tv, sizeof(struct timeval));
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		retval = select(fd_max, &rfds, NULL, NULL, &tvl);
		if (retval) {
			if( FD_ISSET(fd, &rfds) ) {
			    retval = read(fd, buf+l, count - l);
			    l += retval;
			}
		}else{
			return -1;
		}
	}
	
#ifdef DEBUG_MODE
	if (verbose_flag & PINFO) {
		int __i;
		PVERB(PINFO, "# modbus.c read <- ");
		for(__i=0;__i<l;__i++)
			PVERB(PINFO, "%02x ", 
			      *(((u_char *)buf)+__i) );
		PVERB(PINFO, "\n");
	}
#endif	

	return l;
}




int modbus_open(char *name)
{
	int fd;
	struct termios tio;

#ifdef DEMO_MODE 
	return 2;
#endif
	/* Open Serial Port : 8 bits data, 1 bit stop, Even parity, 9600
           baud */
	if ((fd = open(name, O_RDWR | O_NOCTTY)) < 0 ){
		SYSLOG(LOG_ERR, "Cannot open serial port (%s).", name);
		return -1;
	}

	memset(&tio, 0, sizeof(tio));
	tio.c_cflag &= ~CRTSCTS ;
	/* Zero out the baud rate portion of c_cflag */
	tio.c_cflag &= ~CBAUD;
	/* Set baud rate to 9600 */
	tio.c_cflag |= B9600;
	tio.c_cflag &= ~(CSIZE|PARENB|PARODD);
	tio.c_cflag |= CS8 ;
	//tio.c_cflag |= PARENB |PARODD ;
	/* Ignore modem control lines */
	tio.c_cflag |= CLOCAL;
	/* Enable receiver */
	tio.c_cflag |= CREAD;

	/* Turn of Canonical Processing and echo */
	tio.c_lflag &= ~ICANON;
	tio.c_lflag &= ~ECHO;
	/* Set MIN to one and TIME to zero */
	tio.c_cc[VMIN]  = 1;
	tio.c_cc[VTIME] = 0;

	/* Ignore BREAK condition on input */
	tio.c_iflag |= IGNBRK;
	/* ignore framing errors and parity errors */
	tio.c_iflag |=IGNPAR;
	/* Disable NL to CR mapping */
	tio.c_iflag &= ~ICRNL;
	/* Disable XON/XOFF flow control on output and input */
	tio.c_iflag &= ~(IXON|IXOFF);

	/* now clean the modem line and activate the settings for modem */
	tcflush(fd, TCOFLUSH|TCIFLUSH);
	tcsetattr(fd, TCSANOW, &tio);
	
	return fd;
}


int modbus_close(int fd)
{
#ifndef DEMO_MODE
	close(fd);
#endif
	return 0;
}


static unsigned int pk_calc_crc(unsigned char *packet, int len)
{
	unsigned int crc,i,j,carry_flag,a;
	crc = 0xffff;
	for (i=0; i < len; i++) {
		crc = crc ^ (unsigned char) packet[i];
		for (j=0;j<8;j++) {
			a = crc;
			carry_flag = a & 0x0001;
			crc = crc >> 1;
			if (carry_flag == 1)
				crc = crc ^ 0xa001;
		}
	}

	return crc;
}


static unsigned int pk_add_crc(unsigned char *packet, int len)
{
	unsigned int crc;

	crc = pk_calc_crc(packet, len);
	packet[len+1] = crc >> 8;
	packet[len] = crc & 0xff;

	return crc;
}


static int pk_test_crc(unsigned char *packet, int len)
{	
	unsigned int crc;

	crc = pk_calc_crc(packet, len);
	if ( packet[len+1] == (crc >> 8) &&
	     packet[len] == (crc & 0xff)   ) {
		return 1;
	}else{
		return 0;
	}
}


static int modbus_clean_fd(int fd)
{
	char c;
	struct timeval tv;
       
	tv.tv_sec = 0;
	tv.tv_usec = 1000;
	while( soft_read_time_out(tv, fd, &c, 1) == 1);

	return 0;
}


int modbus_read_pk(int fd, unsigned char *packet, int len)
{
	struct timeval tv;
	int len_read;

	tv.tv_sec = 1;
	tv.tv_usec = 0;
	if ( (len_read = soft_read_time_out(tv, fd, packet, len)) < 0) {
		SYSLOG(LOG_ERR, "modbus (%d) time out.", fd);
		return MODBUS_TIME_OUT;
	}
	if (pk_test_crc(packet, len_read - 2) != 1) {
		modbus_clean_fd(fd);
		return MODBUS_BAD_CRC;
	}

	return len_read;
}


int modbus_write_pk(int fd, unsigned char *packet, int len)
{
	struct timeval tv;
	fd_set wfds;
	int fd_max = fd + 1;
	int retval;

	tv.tv_sec = 1;
	tv.tv_usec = 0;
	FD_ZERO(&wfds);
	FD_SET(fd, &wfds);
	retval = select(fd_max, NULL, &wfds, NULL, &tv);
	if (retval) {
		if( FD_ISSET(fd, &wfds) ) {
			retval = write(fd, packet, len);
		}
	}else{
		retval = MODBUS_TIME_OUT;
	}

	return retval;
}


int modbus_make_pk(unsigned char *packet, 
		   int slave, int function, int address, int nor)
{
	int len = 0;

	packet[0] = (unsigned char) slave; /* Slave Address */
	packet[1] = (unsigned char) function; /* Function */
	switch(function) {

		//case 0x02:
	case 0x03:
		//case 0x04:
		//case 0x05:
		//case 0x08:
		packet[2] = (address >> 8) & 0xff; /* Starting Address Hi */
		packet[3] = address & 0xff;        /* Starting Address Lo */
		packet[4] = (nor >> 8) & 0xff;
		packet[5] = nor & 0xff;
		len = 6;
		break;
	default:
		PVERB(PERR,"modbus: function: 0x%x ignored.", function);
		len = 0;
		break;
	}
	if( len ) {
		pk_add_crc(packet, len);
		len += 2;
	}

	return len;
}

		
#if 0
int main(void)
{
	int fd;
	unsigned char pk[MODBUS_PACKET_LENGTH];
	int i,j;
	int len;

	verbose_flag = 0x3;
	fd = modbus_open("/dev/ttyS0");
	memset(pk, 0, MODBUS_PACKET_LENGTH);
	while(1){
		len = modbus_make_pk(pk, 1, 3, 0, 3);
		modbus_write_pk(fd, pk, len);
		len = modbus_read_pk(fd, pk, 3 + (3) * 2 + 2); 
		if(len > 0) {
			for(i=0;i<len;i++)
				printf("0x%02x ",
				       (unsigned char) pk[i]);
			printf("\r");
			fflush(stdout);
		}
	}
	modbus_close(fd);

	return 0;

}

#endif
